# SPDX-License-Identifier: WTFPL
# SPDX-FileCopyrightText: 2023-2024 Anna <cyber@sysrq.in>
# No warranty

from copy import deepcopy
from pathlib import Path

import pytest
from xmldiff.formatting import DiffFormatter
from xmldiff.main import diff_trees

from gentle.generators import AbstractGenerator
from gentle.metadata import MetadataXML


def compare_mxml(old: MetadataXML, new: MetadataXML) -> str:
    return diff_trees(old.xml, new.xml, formatter=DiffFormatter())


@pytest.mark.usefixtures("mxml")
class BaseTestGenerator:
    generator_cls: type[AbstractGenerator]
    generator_data_dir: Path

    def test_pkg_none(self, mxml: MetadataXML):
        self._test_pkg_none(mxml, "pkg_none")

    def test_pkg_empty(self, mxml: MetadataXML):
        self._test_pkg_empty(mxml, "pkg_empty")

    def _test_pkg_none(self, mxml: MetadataXML, dirname: str):
        gen = self.generator_cls(self.generator_data_dir / dirname)
        assert not gen.active

    def _test_pkg_empty(self, mxml: MetadataXML, dirname: str):
        gen = self.generator_cls(self.generator_data_dir / dirname)
        assert gen.active

        mxml_old = deepcopy(mxml)
        gen.update_metadata_xml(mxml)
        assert compare_mxml(mxml_old, mxml) == ""

    def _test_pkg(self, mxml: MetadataXML, dirname: str):
        gen = self.generator_cls(self.generator_data_dir / dirname)
        assert gen.active

        gen.update_metadata_xml(mxml)
        with open(self.generator_data_dir / dirname / "metadata.xml") as file:
            assert mxml.dumps() == file.read().rstrip()

        mxml_prev = deepcopy(mxml)
        gen.update_metadata_xml(mxml)
        assert compare_mxml(mxml_prev, mxml) == ""



# SPDX-License-Identifier: WTFPL
# SPDX-FileCopyrightText: 2023 Anna <cyber@sysrq.in>
# No warranty

"""
Metadata XML generator for Python PEP 621 (pyproject.toml).

The following attributes are supported:

* Upstream maintainer(s)
* Upstream bug tracker
* Upstream changelog
* Upstream documentation
* Remote ID
"""

import logging
from pathlib import Path
import tomllib

from gentle.generators import AbstractGenerator
from gentle.generators.python import (
    BUG_TRACKER_LABELS,
    CHANGELOG_LABELS,
    DOCS_LABELS,
    HOME_REPO_LABELS
)
from gentle.metadata import MetadataXML
from gentle.metadata.types import Person
from gentle.metadata.utils import extract_remote_id

logger = logging.getLogger("pyproject")


class PyprojectGenerator(AbstractGenerator):
    def __init__(self, srcdir: Path):
        self.pyproject_toml = srcdir / "pyproject.toml"

    def update_metadata_xml(self, mxml: MetadataXML) -> None:
        with open(self.pyproject_toml, "rb") as file:
            pyproject = tomllib.load(file)

        if (project := pyproject.get("project")) is None:
            return

        maint_key = "maintainers"
        if maint_key not in project:
            maint_key = "authors"

        for maint in project.get(maint_key, {}):
            person = Person(name=maint.get("name", ""), email=maint.get("email", ""))
            logger.info("Found upstream maintainer: %s", person)
            if not person.name:
                continue
            mxml.add_upstream_maintainer(person)

        for name, value in project.get("urls", {}).items():
            logger.info("Found %s: %s", name, value)
            if name.lower() in BUG_TRACKER_LABELS:
                mxml.set_upstream_bugs_to(value)
            elif name.lower() in CHANGELOG_LABELS:
                mxml.set_upstream_changelog(value)
            elif name.lower() in DOCS_LABELS:
                mxml.set_upstream_doc(value)
            elif name.lower() in HOME_REPO_LABELS:
                if (remote_id := extract_remote_id(value)) is not None:
                    mxml.add_upstream_remote_id(remote_id)

    @property
    def active(self) -> bool:
        return self.pyproject_toml.is_file()



# SPDX-License-Identifier: WTFPL
# SPDX-FileCopyrightText: 2022 Anna <cyber@sysrq.in>
# No warranty

"""
Metadata XML generator for Cargo.

The following attributes are supported:

* Upstream maintainer(s)
* Upstream documentation
* Remote ID
"""

import logging
import tomllib
from pathlib import Path

from gentle.generators import AbstractGenerator
from gentle.metadata import MetadataXML
from gentle.metadata.utils import extract_name_email, extract_remote_id

logger = logging.getLogger("cargo")


class CargoGenerator(AbstractGenerator):
    def __init__(self, srcdir: Path):
        self.srcdir = srcdir
        self.cargo_toml = srcdir / "Cargo.toml"

    def process_metadata(self, package: dict, mxml: MetadataXML) -> None:
        if isinstance(authors := package.get("authors"), list):
            for author in map(extract_name_email, authors):
                if author is None:
                    continue
                logger.info("Found upstream maintainer: %s", author)
                mxml.add_upstream_maintainer(author)

        if isinstance(doc := package.get("documentation"), str):
            logger.info("Found documentation: %s", doc)
            mxml.set_upstream_doc(doc)

        if isinstance(homepage := package.get("homepage"), str):
            logger.info("Found homepage: %s", homepage)
            if (remote_id := extract_remote_id(homepage)) is not None:
                mxml.add_upstream_remote_id(remote_id)

        if isinstance(repo := package.get("repository"), str):
            logger.info("Found repository: %s", repo)
            if (remote_id := extract_remote_id(repo)) is not None:
                mxml.add_upstream_remote_id(remote_id)

    def update_metadata_xml(self, mxml: MetadataXML) -> None:
        with open(self.cargo_toml, "rb") as file:
            crate = tomllib.load(file)

        if (package := crate.get("package")):
            self.process_metadata(package, mxml)

        if (workspace := crate.get("workspace")) is not None:
            if (workspace_package := workspace.get("package")) is not None:
                self.process_metadata(workspace_package, mxml)

            members = set(workspace.get("members", []))
            members -= frozenset(workspace.get("exclude", []))
            for member in members:
                logger.info("Processing workspace member: %s", member)
                member_toml = self.srcdir / member / "Cargo.toml"
                try:
                    with open(member_toml, "rb") as file:
                        member_crate = tomllib.load(file)
                except OSError:
                    continue

                if (member_package := member_crate.get("package")) is not None:
                    self.process_metadata(member_package, mxml)

    @property
    def active(self) -> bool:
        return self.cargo_toml.is_file()



# SPDX-License-Identifier: WTFPL
# SPDX-FileCopyrightText: 2023 Anna <cyber@sysrq.in>
# No warranty

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'gentle'
author = 'Anna <cyber@sysrq.in>'
copyright = f'2022-2024, {author}'
release = '1.0.2'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autosummary',
    'sphinx.ext.autodoc',
    'sphinx.ext.extlinks',
    'sphinx.ext.intersphinx',
    'sphinx_prompt',
]

try:
    import notfound.extension
    extensions.append('notfound.extension')

    notfound_urls_prefix = None
except ModuleNotFoundError:
    pass

try:
    import sphinx_sitemap
    extensions.append('sphinx_sitemap')

    sitemap_locales = [None]
    sitemap_url_scheme = '{link}'
    sitemap_excludes = [
        '404.html',
    ]
except ModuleNotFoundError:
    pass

root_doc = 'toc'
templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']
manpages_url = 'https://docs.sysrq.in/{path}'

extlinks = {
    'bug': ('https://bugs.sysrq.in/show_bug.cgi?id=%s', 'bug #%s'),
    'gentoobug': ('https://bugs.gentoo.org/%s', 'Gentoo bug #%s'),
    'gitweb': (f'https://git.sysrq.in/{project}/tree/%s', '%s'),
    'pypi': ('https://pypi.org/project/%s/', '%s'),
}
intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
}

autosummary_generate = False
autodoc_default_options = {
    'show-inheritance': True,
    'undoc-members': True,
    'member-order': 'bysource',
}

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'insipid'
html_permalinks_icon = '#'
html_theme_options = {
    'globaltoc_maxdepth': 3,
    'right_buttons': ['git-button.html'],
}
html_sidebars = {
    '**': [
        'globaltoc.html',
    ]
}
html_context = {
    'git_repo_url': 'https://git.sysrq.in/gentle/about/',
}

html_static_path = ['_static']
html_title = f'{project} {release}'
html_baseurl = 'https://gentle.sysrq.in/'



from setuptools import setup, find_packages, Command
import io
from os import path


class Test(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        import subprocess
        import sys
        errno = subprocess.call([
            sys.executable, 'commonmark/tests/run_spec_tests.py'])
        raise SystemExit(errno)


tests_require = [
    'flake8==3.9.2',
    'hypothesis==4.24.4',
]


this_directory = path.abspath(path.dirname(__file__))
with io.open(path.join(this_directory, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name="commonmark",
    packages=find_packages(exclude=['tests']),
    version="0.9.1",
    license="BSD-3-Clause",
    description="Python parser for the CommonMark Markdown spec",
    long_description=long_description,
    long_description_content_type='text/x-rst',
    author="Bibek Kafle <bkafle662@gmail.com>, " +
    "Roland Shoemaker <rolandshoemaker@gmail.com>",
    author_email="rolandshoemaker@gmail.com",
    maintainer="Nikolas Nyby",
    maintainer_email="nikolas@gnu.org",
    url="https://github.com/rtfd/commonmark.py",
    keywords=["markup", "markdown", "commonmark"],
    entry_points={
        'console_scripts': [
            'cmark = commonmark.cmark:main',
        ]
    },
    cmdclass={'test': Test},
    install_requires=[
        'future>=0.14.0;python_version<"3"',
    ],
    tests_require=tests_require,
    extras_require={'test': tests_require},
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "Development Status :: 5 - Production/Stable",
        "Environment :: Other Environment",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Topic :: Documentation",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Software Development :: Documentation",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Text Processing :: Markup",
        "Topic :: Utilities"])



# SPDX-License-Identifier: WTFPL
# SPDX-FileCopyrightText: 2023-2024 Anna <cyber@sysrq.in>
# No warranty

from pathlib import Path

import pytest

from gentle.generators.bower import BowerGenerator
from gentle.metadata import MetadataXML
from tests.utils import BaseTestGenerator


class TestBowerGenerator(BaseTestGenerator):
    generator_cls = BowerGenerator
    generator_data_dir = Path(__file__).parent

    @pytest.mark.parametrize("dirname", ["aurelia"])
    def test_pkg(self, mxml: MetadataXML, dirname: str):
        self._test_pkg(mxml, dirname)



# SPDX-License-Identifier: WTFPL
# SPDX-FileCopyrightText: 2023 Anna <cyber@sysrq.in>
# No warranty

"""
Metadata processing routines.
"""

import logging
from pathlib import Path
from typing import Callable

import lxml.etree as ET

from gentle.metadata.types import Person, RemoteID, Upstream

logger = logging.getLogger("metadata")


class MetadataXML:
    """
    Modify :file:`metadata.xml` files.
    """

    def __init__(self, xmlfile: Path, parser: Callable[[Path], Upstream]):
        """
        :param xmlfile: path to the :file:`metadata.xml` file
        :param upstream: pre-parsed :class:`Upstream` object
        """

        self.xmlfile: Path = xmlfile
        self.xml: ET._ElementTree = ET.parse(self.xmlfile)

        self.upstream: Upstream = parser(xmlfile)
        self.modified: bool = False

        self._whitespace_map = {
            " ": "%20",
            "\n": "%0D",
            "\t": "%09",
        }

    def _encode_whitespace(self, url: str) -> str:
        for char, code in self._whitespace_map.items():
            url = url.replace(char, code)
        return url

    def dump(self) -> None:
        """
        Write :file:`metadata.xml` file.
        """

        logger.info("Writing metadata.xml")
        ET.indent(self.xml, space="\t", level=0)
        self.xml.write(self.xmlfile,
                       xml_declaration=True,
                       pretty_print=True,
                       encoding="UTF-8")

    def dumps(self) -> str:
        """
        Convert the object to text.

        :returns: XML data as text
        """

        ET.indent(self.xml, space="\t", level=0)
        return ET.tostring(self.xml.getroot(), encoding="unicode")

    def add_upstream_maintainer(self, person: Person) -> None:
        """
        Add a person to the list of upstream maintainers.

        :param person: upstrem maintainer
        """

        if person in self.upstream.maintainers:
            return

        logger.info("Adding upstream maintainer: %s", person)
        self.upstream.maintainers.append(person)
        upstream = self._make_upstream_element()
        upstream.append(person.to_xml())

        self.modified = True

    def add_upstream_remote_id(self, remote_id: RemoteID) -> None:
        """
        Add an item to the list of remote ids.

        :param remote_id: new remote id
        """

        for old_remote_id in self.upstream.remote_ids:
            if remote_id.attr == old_remote_id.attr:
                return

        logger.info("Adding remote id: %s", remote_id)
        self.upstream.remote_ids.append(remote_id)
        upstream = self._make_upstream_element()
        upstream.append(remote_id.to_xml())

        self.modified = True

    def set_upstream_bugs_to(self, url: str) -> None:
        """
        Set upstream bugs-to URL.

        :param url: new URL
        """

        url = self._encode_whitespace(url)
        if self.upstream.bugs_to:
            return

        logger.info("Setting upstream bug tracker to %s", url)
        self.upstream.bugs_to = url

        upstream = self._make_upstream_element()
        bugs_to = ET.SubElement(upstream, "bugs-to")
        bugs_to.text = url

        self.modified = True

    def set_upstream_changelog(self, url: str) -> None:
        """
        Set upstream changelog URL.

        :param url: new URL
        """

        url = self._encode_whitespace(url)
        if self.upstream.changelog:
            return

        logger.info("Setting upstream changelog to %s", url)
        self.upstream.changelog = url

        upstream = self._make_upstream_element()
        changelog = ET.SubElement(upstream, "changelog")
        changelog.text = url

        self.modified = True

    def set_upstream_doc(self, url: str) -> None:
        """
        Set upstream documentation URL.

        :param url: new URL
        """

        url = self._encode_whitespace(url)
        if self.upstream.doc:
            return

        logger.info("Setting upstream documentation to %s", url)
        self.upstream.doc = url

        upstream = self._make_upstream_element()
        doc = ET.SubElement(upstream, "doc")
        doc.text = url

        self.modified = True

    def _make_upstream_element(self) -> ET._Element:
        if (upstream := self.xml.find("upstream")) is None:
            pkgmetadata = self.xml.getroot()
            upstream = ET.SubElement(pkgmetadata, "upstream")
        return upstream



# SPDX-License-Identifier: WTFPL
# SPDX-FileCopyrightText: 2023 Anna <cyber@sysrq.in>
# No warranty

"""
Types for working with Gentoo package metadata.
"""

from dataclasses import dataclass, field
from enum import Enum, auto

import lxml.etree as ET


class MaintainerStatus(Enum):
    """
    Maintainer status.
    """

    #: Not specified.
    NONE = auto()

    #: Active.
    ACTIVE = auto()

    #: Inactive.
    INACTIVE = auto()


@dataclass
class Person:
    """
    Representation of a person.
    """

    #: Maintainer name.
    name: str = field(default="", compare=False)

    #: Maintainer email.
    email: str = ""

    #: Maintainer activity status.
    status: MaintainerStatus = MaintainerStatus.NONE

    def __str__(self) -> str:
        if self.name and self.email:
            return f"{self.name} <{self.email}>"

        if self.email:
            return self.email

        return self.name

    def to_xml(self, attrib: dict | None = None) -> ET._Element:
        """
        Make an XML ``<maintainer>`` tag.

        :param attrib: attributes for the tag

        :return: :file:`metadata.xml` respresentation of a person
        """

        attrib = attrib or {}
        match self.status:
            case MaintainerStatus.ACTIVE:
                attrib["status"] = "active"
            case MaintainerStatus.INACTIVE:
                attrib["status"] = "inactive"

        result = ET.Element("maintainer", attrib=attrib)
        if self.name:
            name_elem = ET.SubElement(result, "name")
            name_elem.text = self.name
        if self.email:
            email_elem = ET.SubElement(result, "email")
            email_elem.text = self.email

        return result


@dataclass
class RemoteID:
    """
    Representation of a Remote ID.
    """

    #: Site name.
    attr: str

    #: Package identificator on the site.
    value: str

    def __str__(self) -> str:
        return f"{self.attr}:{self.value}"

    def to_xml(self) -> ET._Element:
        """
        Make an XML ``<remote-id>`` tag.

        :return: :file:`metadata.xml` respresentation of a remote id
        """

        remote_elem = ET.Element("remote-id", type=self.attr)
        remote_elem.text = self.value
        return remote_elem


@dataclass
class Upstream:
    """
    Representation of upstream metadata.
    """

    #: Upstream maintainers.
    maintainers: list[Person] = field(default_factory=list)

    #: Upstream changelog.
    changelog: str | None = None

    #: Upstream documentation.
    doc: str | None = None

    #: Upstream bug tracker.
    bugs_to: str | None = None

    #: Upstream identificators on third-party sites.
    remote_ids: list[RemoteID] = field(default_factory=list)



# SPDX-License-Identifier: WTFPL
# SPDX-FileCopyrightText: 2023-2025 Anna <cyber@sysrq.in>
# No warranty

"""
Metadata XML generator for Python Setuptools (setup.cfg).

The following attributes are supported:

* Upstream bug tracker
* Upstream changelog
* Upstream documentation
* Remote ID
"""

import configparser
import logging
from pathlib import Path

from gentle.generators import AbstractGenerator
from gentle.generators.python import (
    BUG_TRACKER_LABELS,
    CHANGELOG_LABELS,
    DOCS_LABELS,
    HOME_REPO_LABELS
)
from gentle.metadata import MetadataXML
from gentle.metadata.utils import extract_remote_id

logger = logging.getLogger("setuptools")


class SetuptoolsGenerator(AbstractGenerator):
    def __init__(self, srcdir: Path):
        self.setup_cfg = srcdir / "setup.cfg"

    def update_metadata_xml(self, mxml: MetadataXML) -> None:
        cfg = configparser.ConfigParser()
        try:
            cfg.read(self.setup_cfg)
        except configparser.Error:
            return

        if not cfg.has_section("metadata"):
            return
        metadata = cfg["metadata"]

        # Not parsing "author" and "maintainer" because of uncertain format

        if (url := metadata.get("url")) is None:
            url = metadata.get("home-page")

        if url is not None:
            if (remote_id := extract_remote_id(url)) is not None:
                logger.info("Found homepage: %s", url)
                mxml.add_upstream_remote_id(remote_id)

        for key_val_str in metadata.get("project_urls", "").strip().split("\n"):
            key_val = [item.strip() for item in key_val_str.split("=", maxsplit=1)]
            if len(key_val) != 2:
                continue

            name = key_val[0].lower()
            value = key_val[1]
            logger.info("Found %s: %s", name, value)
            if name in BUG_TRACKER_LABELS:
                mxml.set_upstream_bugs_to(value)
            elif name in CHANGELOG_LABELS:
                mxml.set_upstream_changelog(value)
            elif name in DOCS_LABELS:
                mxml.set_upstream_doc(value)
            elif name in HOME_REPO_LABELS:
                if (remote_id := extract_remote_id(value)) is not None:
                    mxml.add_upstream_remote_id(remote_id)

    @property
    def active(self) -> bool:
        return self.setup_cfg.is_file()



# SPDX-License-Identifier: WTFPL
# SPDX-FileCopyrightText: 2023 Anna <cyber@sysrq.in>
# No warranty

"""
Common data for Python generators.
"""

#: Project URL names matching upstream bug tracker.
BUG_TRACKER_LABELS = [
    "bug tracker",
    "bug reports",
    "bugtracker",
    "bugs",
    "issue tracker",
    "issues",
    "tracker",
    "github: issues",
]

#: Project URL names matching upstream changelog.
CHANGELOG_LABELS = [
    "changelog",
    "changes",
    "release notes",
]

#: Project URL names matching upstream documentation.
DOCS_LABELS = [
    "doc",
    "docs",
    "documentation",
    "docs: rtd",
]

#: Project URL names matching upstream repository or homepage.
HOME_REPO_LABELS = [
    "code",
    "source",
    "source code",
    "repo",
    "repository",
    "home",
    "homepage",
    "github",
    "github: repo",
]



# SPDX-License-Identifier: WTFPL
# SPDX-FileCopyrightText: 2023-2024 Anna <cyber@sysrq.in>
# No warranty

from pathlib import Path

from gentle.generators.pubspec import PubspecGenerator
from tests.utils import BaseTestGenerator


class TestPubspecGenerator(BaseTestGenerator):
    generator_cls = PubspecGenerator
    generator_data_dir = Path(__file__).parent



# SPDX-License-Identifier: WTFPL
# SPDX-FileCopyrightText: 2023 Anna <cyber@sysrq.in>
# No warranty

"""
Metadata XML generator for Bower.

The following attributes are supported:

* Upstream maintainer(s)
* Remote ID
"""

import json
import logging
from pathlib import Path

from gentle.generators import AbstractGenerator
from gentle.metadata import MetadataXML
from gentle.metadata.types import Person
from gentle.metadata.utils import extract_name_email, extract_remote_id

logger = logging.getLogger("bower")


class BowerGenerator(AbstractGenerator):
    def __init__(self, srcdir: Path):
        self.bower_json = srcdir / "bower.json"

    def update_metadata_xml(self, mxml: MetadataXML) -> None:
        with open(self.bower_json) as file:
            try:
                package = json.load(file)
            except json.decoder.JSONDecodeError:
                return

        for maint in package.get("authors", []):
            logger.info("Found upstream maintainer: %s", maint)
            if isinstance(maint, str):
                person = extract_name_email(maint)
            else:
                person = Person(name=maint.get("name", ""),
                                email=maint.get("email", ""))

            if person is not None and person.name:
                mxml.add_upstream_maintainer(person)

        if (homepage := package.get("homepage")) is not None:
            logger.info("Found homepage: %s", homepage)
            if (remote_id := extract_remote_id(homepage)) is not None:
                mxml.add_upstream_remote_id(remote_id)

    @property
    def active(self) -> bool:
        return self.bower_json.is_file()



# SPDX-License-Identifier: WTFPL
# SPDX-FileCopyrightText: 2023 Anna <cyber@sysrq.in>
# No warranty

"""
Metadata XML generator for Python PEP 517.

The following attributes are supported:

* Upstream maintainer(s)
* Upstream bug tracker
* Upstream changelog
* Upstream documentation
* Remote ID
"""

import logging
import shutil
from importlib.metadata import PathDistribution, PackageMetadata
from pathlib import Path
from tempfile import TemporaryDirectory

from gentle.generators import AbstractGenerator
from gentle.generators.python import (
    BUG_TRACKER_LABELS,
    CHANGELOG_LABELS,
    DOCS_LABELS,
    HOME_REPO_LABELS
)
from gentle.metadata import MetadataXML
from gentle.metadata.utils import extract_name_email, extract_remote_id

try:
    import pyproject_hooks
    from build import BuildException, BuildBackendException, ProjectBuilder
    from build.env import DefaultIsolatedEnv
    _HAS_BUILD = True
except ModuleNotFoundError:
    _HAS_BUILD = False

logger = logging.getLogger("wheel")


class WheelGenerator(AbstractGenerator):
    def __init__(self, srcdir: Path):
        self.srcdir = srcdir

        self.pyproject_toml = srcdir / "pyproject.toml"
        self.setup_cfg = srcdir / "setup.cfg"
        self.setup_py = srcdir / "setup.py"

        self.pkg_info = srcdir / "PKG-INFO"

    def project_wheel_metadata(self) -> PackageMetadata:
        with DefaultIsolatedEnv(installer="uv") as env:
            builder = ProjectBuilder.from_isolated_env(
                env, self.srcdir,
                runner=pyproject_hooks.quiet_subprocess_runner,
            )
            env.install(builder.build_system_requires)
            env.install(builder.get_requires_for_build("wheel"))
            with TemporaryDirectory(prefix="gentle-wheel-") as tmpdir:
                path = Path(builder.metadata_path(tmpdir))
                return PathDistribution(path).metadata

    # pylint: disable=too-many-branches
    def update_metadata_xml(self, mxml: MetadataXML) -> None:
        try:
            metadata: PackageMetadata = self.project_wheel_metadata()
        except (BuildException, BuildBackendException):
            return

        maint_key = "maintainer"
        if maint_key not in metadata:
            maint_key = "author"

        maint_email_key = "maintainer-email"
        if maint_email_key not in metadata:
            maint_email_key = "author-email"

        maintainers = []
        for key in (maint_key, maint_email_key):
            if key not in metadata:
                continue
            maintainers += [entry.strip() for entry in metadata[key].split(",")]

        for maint in map(extract_name_email, maintainers):
            if maint is None:
                continue
            logger.info("Found upstream maintainer: %s", maint)
            mxml.add_upstream_maintainer(maint)

        if (homepage := metadata.get("home-page")) is not None:  # type: ignore
            logger.info("Found homepage: %s", homepage)
            if (remote_id := extract_remote_id(homepage)) is not None:
                mxml.add_upstream_remote_id(remote_id)

        for entry in metadata.get_all("project-url", []):
            name, value = [item.strip()
                           for item in entry.split(",", maxsplit=1)]
            logger.info("Found %s: %s", name, value)
            if name.lower() in BUG_TRACKER_LABELS:
                mxml.set_upstream_bugs_to(value)
            elif name.lower() in CHANGELOG_LABELS:
                mxml.set_upstream_changelog(value)
            elif name.lower() in DOCS_LABELS:
                mxml.set_upstream_doc(value)
            elif name.lower() in HOME_REPO_LABELS:
                if (remote_id := extract_remote_id(value)) is not None:
                    mxml.add_upstream_remote_id(remote_id)

    @property
    def active(self) -> bool:
        return (
            _HAS_BUILD
            and not self.pkg_info.is_file()
            and shutil.which("uv") is not None
            and (
                self.pyproject_toml.is_file()
                or self.setup_cfg.is_file()
                or self.setup_py.is_file()
            )
        )

    @property
    def slow(self) -> bool:
        return True



# SPDX-License-Identifier: WTFPL
# SPDX-FileCopyrightText: 2023 Anna <cyber@sysrq.in>
# No warranty

"""
Generic generator interface.
"""

from abc import ABC, abstractmethod
from pathlib import Path

from gentle.metadata import MetadataXML


class AbstractGenerator(ABC):
    """
    Generic class for metadata generators.
    """

    @abstractmethod
    def __init__(self, srcdir: Path):
        """
        :param srcdir: path to unpacked sources
        """

    @abstractmethod
    def update_metadata_xml(self, mxml: MetadataXML) -> None:
        """
        Update metadata object in place.

        :param mxml: :file:`metadata.xml` object
        """

    @property
    @abstractmethod
    def active(self) -> bool:
        """
        Whether generator works.
        """

    @property
    def slow(self) -> bool:
        """
        Whether generator takes long time to finish.
        """

        return False



# SPDX-License-Identifier: WTFPL
# SPDX-FileCopyrightText: 2022 Anna <cyber@sysrq.in>
# No warranty

"""
Metadata XML generator for Ruby Gems.

The following attributes are supported:

* Upstream bug tracker
* Upstream changelog
* Upstream documentation
* Remote ID
"""

import logging
import shutil
import subprocess
from pathlib import Path

from gentle.generators import AbstractGenerator
from gentle.metadata import MetadataXML
from gentle.metadata.utils import extract_remote_id

try:
    import yaml
    from yaml import CSafeLoader
    _HAS_PYYAML = True

    class VersionTag(yaml.YAMLObject):
        """ Dummy version tag """
        yaml_tag = "!ruby/object:Gem::Version"
        yaml_loader = CSafeLoader

    class RequirementTag(yaml.YAMLObject):
        """ Dummy requirement tag """
        yaml_tag = "!ruby/object:Gem::Requirement"
        yaml_loader = CSafeLoader

    class DependencyTag(yaml.YAMLObject):
        """ Dummy dependency tag """
        yaml_tag = "!ruby/object:Gem::Dependency"
        yaml_loader = CSafeLoader

    class SpecificationTag(yaml.YAMLObject):
        """ Dummy specification tag """
        yaml_tag = "!ruby/object:Gem::Specification"
        yaml_loader = CSafeLoader
except ModuleNotFoundError:
    _HAS_PYYAML = False

logger = logging.getLogger("gemspec")


class GemspecGenerator(AbstractGenerator):
    def __init__(self, srcdir: Path):
        self.srcdir = srcdir

        self.gemspec_files = list(srcdir.glob("*.gemspec"))
        self.metadata_yml = srcdir / "all" / "metadata"

        self.ruby = shutil.which("ruby")

    def update_metadata_xml(self, mxml: MetadataXML) -> None:
        if self.metadata_yml.is_file():
            with open(self.metadata_yml, "rb") as file:
                data = file.read()
        else:
            gemspec = self.gemspec_files[0]
            code = "print Gem::Specification.load(gets.chomp).to_yaml"
            data = subprocess.run([str(self.ruby), "-e", code],  # nosec B603
                                  input=str(gemspec).encode(),
                                  cwd=self.srcdir,
                                  check=False,
                                  capture_output=True).stdout

        if (metadata := yaml.load(data, CSafeLoader)) is None:
            return

        if metadata.homepage:
            logger.info("Found homepage: %s", metadata.homepage)
            if (remote_id := extract_remote_id(metadata.homepage)) is not None:
                mxml.add_upstream_remote_id(remote_id)

        for name, value in metadata.metadata.items():
            if not isinstance(value, str):
                continue

            logger.info("Found %s: %s", name, value)
            match name:
                case "bug_tracker_uri":
                    mxml.set_upstream_bugs_to(value)
                case "changelog_uri":
                    mxml.set_upstream_changelog(value)
                case "documentation_uri":
                    mxml.set_upstream_doc(value)
                case "homepage_uri" | "source_code_uri":
                    if (remote_id := extract_remote_id(value)) is not None:
                        mxml.add_upstream_remote_id(remote_id)

    @property
    def active(self) -> bool:
        return (
            _HAS_PYYAML
            and (
                self.metadata_yml.is_file()
                or (
                    len(self.gemspec_files) == 1
                    and self.ruby is not None
                )
            )
        )



# SPDX-License-Identifier: WTFPL
# SPDX-FileCopyrightText: 2023-2024 Anna <cyber@sysrq.in>
# No warranty

from pathlib import Path

import build
import pytest

from gentle.generators.python.wheel import WheelGenerator
from gentle.metadata import MetadataXML
from tests.utils import BaseTestGenerator


class TestWheelGenerator(BaseTestGenerator):
    generator_cls = WheelGenerator
    generator_data_dir = Path(__file__).parent

    @pytest.mark.parametrize("dirname", ["pyproject.toml", "setup.cfg", "setup.py"])
    def test_pkg_empty(self, monkeypatch: pytest.MonkeyPatch, mxml: MetadataXML, dirname: str):
        # subprocess calls can make this test fail
        def blackhole(*args, **kwargs):
            pass
        monkeypatch.setattr(build.env, "run_subprocess", blackhole)

        self._test_pkg_empty(mxml, "pkg_empty/" + dirname)

    @pytest.mark.net
    @pytest.mark.parametrize("dirname", ["rich", "commonmark"])
    def test_pkg(self, mxml: MetadataXML, dirname: str):
        self._test_pkg(mxml, dirname)



# SPDX-License-Identifier: WTFPL
# SPDX-FileCopyrightText: 2023 Anna <cyber@sysrq.in>
# No warranty

"""
Metadata XML generator for Perl CPAN::Meta::Spec.

The following attributes are supported:

* Upstream maintainer(s)
* Upstream bug tracker
* Remote ID
"""

import json
import logging
import shutil
import subprocess
from pathlib import Path

from gentle.generators import AbstractGenerator
from gentle.metadata import MetadataXML
from gentle.metadata.utils import extract_name_email, extract_remote_id

logger = logging.getLogger("cpan")


class CpanGenerator(AbstractGenerator):
    def __init__(self, srcdir: Path):
        self.srcdir = srcdir

        self.meta_json = srcdir / "META.json"
        self.makefile_pl = srcdir / "Makefile.PL"

        self.perl = shutil.which("perl")

    # pylint: disable=too-many-branches
    def update_metadata_xml(self, mxml: MetadataXML) -> None:
        meta_json = self.meta_json
        if not self.meta_json.is_file() and self.makefile_pl.is_file():
            cmd = [str(self.perl), str(self.makefile_pl)]
            subprocess.run(cmd, cwd=self.srcdir, check=False)  # nosec B603

            meta_json = self.srcdir / "MYMETA.json"

        if not meta_json.is_file():
            return

        with open(meta_json) as file:
            try:
                meta = json.load(file)
            except json.decoder.JSONDecodeError:
                return

        # Support Spec 2 only
        if not (
            "meta-spec" in meta
            and isinstance(meta["meta-spec"], dict)
            and str(meta["meta-spec"].get("version")) == "2"
        ):
            return

        for author in map(extract_name_email, meta.get("author", [])):
            if author is None:
                continue
            logger.info("Found upstream maintainer: %s", author)
            mxml.add_upstream_maintainer(author)

        for name, value in meta.get("resources", {}).items():
            match name:
                case "homepage":
                    if not isinstance(value, str):
                        continue

                    logger.info("Found homepage: %s", value)
                    if (remote_id := extract_remote_id(value)) is not None:
                        mxml.add_upstream_remote_id(remote_id)
                case "bugtracker":
                    if not isinstance(value, dict):
                        continue

                    if "web" in value:
                        bugs_to = value["web"]
                    elif "mailto" in value:
                        bugs_to = "mailto:" + value["mailto"]
                    else:
                        continue
                    logger.info("Found bug tracker: %s", bugs_to)
                    mxml.set_upstream_bugs_to(bugs_to)
                case "repository":
                    if not isinstance(value, dict):
                        continue

                    if (repo := value.get("web")) is None:
                        continue
                    logger.info("Found repository: %s", repo)
                    if (remote_id := extract_remote_id(repo)) is not None:
                        mxml.add_upstream_remote_id(remote_id)
                case _:
                    if not isinstance(value, str):
                        continue

                    logger.info("Found %s: %s", name, value)
                    if (remote_id := extract_remote_id(value)) is not None:
                        mxml.add_upstream_remote_id(remote_id)

    @property
    def active(self) -> bool:
        return (
            self.meta_json.is_file()
            or (
                self.perl is not None
                and self.makefile_pl.is_file()
            )
        )

    @property
    def slow(self) -> bool:
        return not self.meta_json.is_file()



# SPDX-License-Identifier: WTFPL
# SPDX-FileCopyrightText: 2022 Anna <cyber@sysrq.in>
# No warranty

"""
Metadata XML generator for Crystal Shards.

The following attributes are supported:

* Upstream maintainer(s)
* Upstream documentation
* Remote ID
"""

import logging
from pathlib import Path

from gentle.generators import AbstractGenerator
from gentle.metadata import MetadataXML
from gentle.metadata.utils import extract_name_email, extract_remote_id

try:
    import yaml
    from yaml import CSafeLoader
    _HAS_PYYAML = True
except ModuleNotFoundError:
    _HAS_PYYAML = False

logger = logging.getLogger("shards")


class ShardsGenerator(AbstractGenerator):
    def __init__(self, srcdir: Path):
        self.shard_yml = srcdir / "shard.yml"

    def update_metadata_xml(self, mxml: MetadataXML) -> None:
        with open(self.shard_yml) as file:
            if (shard := yaml.load(file, CSafeLoader)) is None:
                return

        for author in map(extract_name_email, shard.get("authors", [])):
            if author is None:
                continue
            logger.info("Found upstream maintainer: %s", author)
            mxml.add_upstream_maintainer(author)

        if (doc := shard.get("documentation")) is not None:
            logger.info("Found documentation: %s", doc)
            mxml.set_upstream_doc(doc)

        if (homepage := shard.get("homepage")) is not None:
            logger.info("Found homepage: %s", homepage)
            if (remote_id := extract_remote_id(homepage)) is not None:
                mxml.add_upstream_remote_id(remote_id)

        if (repo := shard.get("repository")) is not None:
            logger.info("Found repository: %s", repo)
            if (remote_id := extract_remote_id(repo)) is not None:
                mxml.add_upstream_remote_id(remote_id)

    @property
    def active(self) -> bool:
        return _HAS_PYYAML and self.shard_yml.is_file()



# SPDX-License-Identifier: WTFPL
# SPDX-FileCopyrightText: 2022-2023 Anna <cyber@sysrq.in>
# No warranty

import argparse
import importlib.metadata
import importlib.util
import logging
import os
import sys
from pathlib import Path
from tempfile import TemporaryDirectory

import gentle
from gentle.generators import AbstractGenerator
from gentle.metadata import MetadataXML

_HAS_PORTAGE = importlib.util.find_spec("portage") is not None

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("main")


def main() -> None:
    """
    Parse command-line arguments and run the program.
    """

    pm = []
    if _HAS_PORTAGE:
        pm.append("portage")

    if len(pm) == 0:
        raise RuntimeError("No package manager installed. Aborting")

    parser = argparse.ArgumentParser("gentle", description=gentle.__doc__)
    parser.add_argument("ebuild", type=Path, help="path to the ebuild file")
    parser.add_argument("--api", "-a", choices=pm, default=pm[0],
                        help="package manager API to use")
    parser.add_argument("--prefix", "-p",
                        help="value of EPREFIX")
    parser.add_argument("--quick", "-q", action="store_true",
                        help="skip slow generators")
    parser.add_argument("-v", action="version", version=gentle.__version__)
    args = parser.parse_args()

    if args.prefix is not None:
        os.environ["PORTAGE_OVERRIDE_EPREFIX"] = args.prefix

    with TemporaryDirectory(prefix="gentle-") as tmpdir:
        # pylint: disable=import-outside-toplevel
        match args.api:
            case "portage":
                from gentle.pms.portagepm import parse_mxml, src_unpack

        mxml_file = args.ebuild.parent / "metadata.xml"
        try:
            mxml = MetadataXML(mxml_file, parse_mxml)
        except FileNotFoundError:
            logger.error("Ebuild's metadata.xml file is missing, create it "
                         "before running gentle")
            sys.exit(1)

        srcdir = src_unpack(args.ebuild, tmpdir)
        for ep in importlib.metadata.entry_points(group="gentle.generators"):
            cls = ep.load()
            if not (isinstance(cls, type)
                    and issubclass(cls, AbstractGenerator)):
                logger.warning("Skipping unsuitable %s", cls.__name__)
                continue

            generator = cls(srcdir)
            if not generator.active:
                continue
            if args.quick and generator.slow:
                continue
            logger.info("Starting %s", cls.__name__)
            generator.update_metadata_xml(mxml)

    if mxml.modified:
        mxml.dump()


if __name__ == "__main__":
    main()



# SPDX-License-Identifier: WTFPL
# SPDX-FileCopyrightText: 2023 Anna <cyber@sysrq.in>
# No warranty

"""
Utility functions and classes.
"""

import lxml.etree as ET


def stringify(element: ET._Element) -> str:
    """
    Extract all text from the given XML element.

    :param element: XML element object

    :returns: text from the given element
    """

    return "".join(
        (text for text in element.itertext() if isinstance(text, str))
    )
